#include <fstream>
#include <iomanip>
#include <mimas/mm_image_funcs.h>
#include <mimas/mm_image_fileoutput.h>
#include <qmessagebox.h>
#include <qscrollview.h>
#include "cameraView.h"

using namespace std;

CameraView::CameraView( QScrollView *_scrollView, const char *name ):
  QGLWidget( _scrollView->viewport(), name ),
  input( MODE_640x480_YUV422, FRAMERATE_15 ), scrollView(_scrollView),
  zoom(1.0), mode(Colour), wheelFrac(0), doThreshold(false),
  thresholdValue(10), grabbing(false), grabCounter(0)
{
  // Start timer.
  startTimer( 0 );
}

CameraView::~CameraView(void)
{
}

void CameraView::refresh(void)
{
  try {
    // Grab image.
    input >> image;

    switch ( mode ) {
    case Colour:
      // No operation.
      break;
    case GrayScale:
      image = mimas::mm_image< unsigned char >( image );
      break;
    case Sobel:
      if ( doThreshold )
        image = mimas::bilevel( mimas::edgeSobel
                                ( mimas::mm_image< int >( image ) ),
                                thresholdValue, 0, 255 );
      else
        image = mimas::normalise( mimas::edgeSobel
                                  ( mimas::mm_image< int >( image ) ),
                                  0, 255 );
      break;
    };    

    // Adapt size of camera-view if necessary.
    adaptSize();

    if ( grabbing & !grabFilePrefix.empty() ) {

      ostringstream fileName;
      fileName << grabFilePrefix << setfill('0') << setw( 4 )
               << grabCounter << ".ppm";

      std::ofstream f( fileName.str().c_str(), std::ios::binary );
      MMERROR( f << image, mimas::mm_exception, ,
               "Couldn't write file \"" << fileName.str() << "\"." );
      

      grabCounter++;
    };

    // Update camera-view.
    repaint();

  } catch ( exception &e ) {

    killTimers();

    if ( QMessageBox::critical( this, "Error",
                                QString( "Error grabbing image: %1\n"
                                         "Exit application?" ).
                                arg( e.what() ),
                                QMessageBox::Yes,
                                QMessageBox::No | QMessageBox::Default ) ==
         QMessageBox::Yes )
      qApp->quit();

    startTimer( 0 );

  };
}

void CameraView::setMode( int _mode )
{
  mode = (Mode)_mode;
  update();
}

void CameraView::setShutter( int value )
{
  input.set_feature_value( FEATURE_SHUTTER, value );
}

void CameraView::setGain( int value )
{
  input.set_feature_value( FEATURE_GAIN, value );
}

void CameraView::setWhiteBalance( int value )
{
  input.set_feature_value( FEATURE_WHITE_BALANCE, value );
}

void CameraView::setZoom( int value )
{
  zoom = 0.001 * value;
  adaptSize();
  update();
}

void CameraView::useThreshold( bool _on )
{
  doThreshold = _on;
}

void CameraView::setThreshold( int _value )
{
  thresholdValue = _value;
}

void CameraView::setGrabbing( bool _grabbing )
{
  grabbing = _grabbing;
}

void CameraView::setGrabFilePrefix(const QString &_prefix )
{
  grabFilePrefix = (const char *)_prefix;
}

void CameraView::initializeGL(void)
{
  qglClearColor( black );
}

void CameraView::resizeGL( int w, int h )
{
  glViewport( 0, 0, (GLint)w, (GLint)h );
}

void CameraView::paintGL(void)
{
  glPixelStorei( GL_UNPACK_ALIGNMENT, 1 );
  glRasterPos2i( -1, 1 );
  glPixelZoom( zoom, -zoom );
  // Direct access to image-memory!
  glDrawPixels( image.getWidth(), image.getHeight(),
                GL_BGRA, GL_UNSIGNED_BYTE, image.rawData().data() );
};

void CameraView::mouseReleaseEvent( QMouseEvent *e )
{
  if ( mousePressed ) {
    emit stopDrag();
    mousePressed = false;
  };
}

void CameraView::mouseMoveEvent( QMouseEvent *e )
{
  if ( mousePressed ) {
    int
      newMouseX = e->x(),
      newMouseY = e->y();
    emit mouseDragged( ( newMouseX - mouseX ) / zoom,
                       ( newMouseY - mouseY ) / zoom );
  };
}

void CameraView::mousePressEvent( QMouseEvent *e )
{
  if ( e->button() == Qt::LeftButton ) {
    emit startDrag( 1 );
    mousePressed = true;
    mouseX = e->x();
    mouseY = e->y();
  } else if ( e->button() == Qt::RightButton ) {
    emit startDrag( 2 );
    mousePressed = true;
    mouseX = e->x();
    mouseY = e->y();
  };
}

void CameraView::timerEvent( QTimerEvent * )
{
  refresh();
}

void CameraView::wheelEvent( QWheelEvent *e )
{
  wheelFrac += e->delta();
  int change = wheelFrac / 120;
  if ( change != 0 ) {
    wheelFrac -= change * 120;
    emit wheelTick( change );
  };
}

void CameraView::adaptSize(void)
{
  int
    w = (int)( image.getWidth() * zoom ),
    h = (int)( image.getHeight() * zoom );
  if ( width() != w || height() != h ) {
    scrollView->resizeContents( w, h );
    resize( w, h );
  };
}
