#include <errno.h>
#include <iostream>
#include <mimas/mm_exception.h>
#include <termios.h>
#include <fcntl.h>
#include "microscope.h"

using namespace LibSerial;
using namespace mimas;
using namespace std;

const string Microscope::IDENTIFICATION = "DMLAM 70 71 72 73 76 77 78 84";

Microscope::Microscope( const string &_device ) throw (mm_exception):
  device(_device)
{
  // Temporary hack to setup the port properly.
  int fd = open( device.c_str(), O_RDWR | O_NOCTTY );
  MMERROR( fd>=0, mm_exception, ,
           "Error opening device \"" << device << "\": "
           << strerror( errno ) );
  
  // Open serial port.
  port.Open( device );
  MMERROR( port.good(), mm_exception, ,
           "Error opening device \"" << device << "\": "
           << strerror( errno ) );

  // Set baud rate.
  port.SetBaudRate( SerialStreamBuf::BAUD_19200 );
  MMERROR( port.good(), mm_exception, ,
           "Error setting baud rate of device \"" << device << "\": "
           << strerror( errno ) );

  // Set number of data bitport.
  port.SetCharSize( SerialStreamBuf::CHAR_SIZE_8 );
  MMERROR( port.good(), mm_exception, ,
           "Error setting character size of device \"" << device << "\": "
           << strerror( errno ) );

  // Use software flow control.
  port.SetFlowControl( SerialStreamBuf::FLOW_CONTROL_SOFT );
  MMERROR( port.good(), mm_exception, ,
           "Error switching device \"" << device
           << "\" to software flow control: " << strerror( errno ) );

  // Timeout of 10 seconds.
  port.SetVTime( 100 );
  MMERROR( port.good(), mm_exception, ,
           "Error setting VTIME of \"" << device
           << "\": " << strerror( errno ) );
  // Non-blocking read.
  port.SetVMin( 0 );
  MMERROR( port.good(), mm_exception, ,
           "Error setting VMIN of \"" << device
           << "\": " << strerror( errno ) );

  // Probe for microscope.
  probe();

  // Initialise drives.
  driveX.init( this, "72034", "72023", "72024", "72022", "72005",
               "72025", "72021" );
  driveY.init( this, "73034", "73023", "73024", "73022", "73005",
               "73025", "73021" );
  driveZ.init( this, "71042", "71023", "71024", "71022", "71005",
               "71025", "71021" );
  
  // Initialise lamp.
  lamp.init( this, "77023", "77024", "77021", "77020" );

  // Initialise nosepiece.
  nosePiece.init( this, "76023", "76024", "76022", "76030 1" );
}

void Microscope::probe(void) throw (mm_exception)
{
  // Testing for microscope.
  string line = readLine( "70001" );
  MMERROR( line == IDENTIFICATION, mm_exception, ,
           "Microscope-identification doesn't match \"" << IDENTIFICATION
           << "\". Got \"" << line << "\" instead." );
}

void Microscope::checkEcho( const string &command ) throw (mm_exception)
{
  port >> ws;
  string echo;
  while ( echo.length() < command.length() ) {
    int c = port.get();
    MMERROR( c != EOF, mm_exception, ,
             "Unexpected end of input while checking for echo \"" << command
             << "\"." );
    echo += (char)c;
  };
  MMERROR( echo == command, mm_exception, ,
           "Expecting echo \"" << command
           << "\", but got \"" << echo << "\" instead." );
}

void Microscope::command( const string &cmd ) throw (mm_exception)
{
  MMERROR( port << cmd << '\r' << flush, mm_exception, ,
           "Error doing output through serial port \""
           << device << "\"." );
  checkEcho( cmd );
}

void Microscope::query( const string &command )
  throw (mm_exception)
{
  MMERROR( port << command << '\r' << flush, mm_exception, ,
           "Error doing output through serial port \""
           << device << "\"." );
  checkEcho( command );
  port >> ws;
}

void Microscope::sendInt( const string &command, int value, bool wait )
  throw (mm_exception)
{
  MMERROR( port << command << ' ' << value << '\r' << flush, mm_exception, ,
           "Error doing output through serial port \""
           << device << "\"." );
  if ( wait ) {
    checkEcho( command );
  };
}

string Microscope::readLine(void) throw (mm_exception)
{
  string line;
  MMERROR( getline( port, line, '\r' ), mm_exception, ,
           "Timeout getting input from serial port \""
           << device << "\"." );
  return line;
}

string Microscope::readLine( const string &command )
  throw (mm_exception)
{
  query( command );
  return readLine();
}

string Microscope::readString(void) throw (mm_exception)
{
  string s; port >> s;
  return s;
}

string Microscope::readString( const string &command ) throw (mm_exception)
{
  query( command );
  return readString();
}

int Microscope::readInt( const string &command )
  throw (mm_exception)
{
  return atoi( readString( command ).c_str() );
}

double Microscope::readDouble( const string &command ) throw (mm_exception)
{
  return atof( readString( command ).c_str() );
}

int Microscope::readChar(void) throw (mm_exception)
{
  return port.get();
}
